# -*- coding: utf-8 -*-
"""Top-level package for Better Dict."""

__author__ = 'Erik Ingwersen'
__email__ = 'erik.ingwersen@br.ey.com'
__version__ = '0.0.1'


from .core import BetterDict
from .utils import *



# file generated by setuptools_scm
# don't change, don't track in version control
__version__ = version = '0.1.dev1+gcc8fd61.d20230330'
__version_tuple__ = version_tuple = (0, 1, 'dev1', 'gcc8fd61.d20230330')



"""
Module defines the :class:`.BetterDict` class.

The :class:`.BetterDict` class is a subclass of the :class:`dict` class that
provides additional functionality to the dictionary. The additional
functionality includes:

- Accessing the dictionary keys by value.
- Manipulating the dictionary keys and values using index notation.
- Accessing and manipulating the dictionary values using dot notation.
- Accessing the dictionary values by their data types.
- Saving and loading the dictionary to and from a file.
- Creating the dictionary from a :class:`pandas.DataFrame` object.
- Creating the dictionary from a :class:`numpy.ndarray` object.
- Creating the dictionary from a :class:`numpy.matrix` object.
- Creating the dictionary from a :class:`pandas.Series` object.
- Applying a function to the dictionary values and keys.
- Finding keys using fuzzy matching.
- Renaming the dictionary keys.

"""
from __future__ import annotations

from abc import abstractmethod
from difflib import get_close_matches
from typing import Any, Hashable, Iterable, List, Tuple

import numpy as np
import pandas as pd

from better_dict.accessors import ILoc, VLoc
from better_dict.io import JoblibIOMixin, PicklerMixin
from better_dict.utils import (flatten, iterable_not_string, make_list,
                               same_length)


def translate_dtype(dtype: str) -> type | Tuple[type, ...]:
    """Translate the ``dtype`` string to the corresponding Python type.

    Parameters
    ----------
    dtype : str
        Data type represented as string to translate.

    Returns
    -------
    type | Tuple[type, ...]
        Type or tuple of types corresponding to the text data types.

    Raises
    ------
    ValueError
        If the data type string is not recognized.
    """
    if dtype == "number":
        return float, int
    if dtype == "string":
        return str
    if dtype == "datetime":
        return pd.Timestamp
    if dtype == "pandas":
        return pd.DataFrame, pd.Series
    if dtype == "numpy":
        return np.ndarray, np.matrix
    raise ValueError(f"Invalid dtype: {dtype!r}")


class ApplyMixin:
    """
    Mixin class that adds the apply method to the ``BetterDict`` class.

    The ``apply`` method enables operations on the values or keys of the
    dictionary.

    """

    @abstractmethod
    def rename(self, keys: dict) -> None:
        """Rename the keys of the dictionary."""

    @abstractmethod
    def keys(self):
        """Get the keys of the dictionary."""

    @abstractmethod
    def values(self):
        """Get the values of the dictionary."""

    @abstractmethod
    def __setitem__(self, key, value):
        """Set the value of the dictionary at the given key."""

    def apply(self, func, *args, axis=1, **kwargs):
        """Apply a function to the ``BetterDict`` keys or values.

        Parameters
        ----------
        func : callable
            Function to apply.
        *args
            Positional arguments to pass to the function.
        **kwargs
            Keyword arguments to pass to the function.
        axis : int, default 1
            Axis along which to apply the function. If ``axis=0``, the function
            is applied to the keys of the dictionary. If ``axis=1``, the
            function is applied to the values of the dictionary.

        Returns
        -------
        BetterDict
            Result of applying the function to the object.

        Notes
        -----
        The class mimics the ``pandas.DataFrame.apply`` method.
        """
        if axis == 1:
            return self.value_apply(func, *args, **kwargs)
        return self.keys_apply(func, *args, **kwargs)

    def keys_apply(self, func, *args, **kwargs):
        """Apply the function to the keys of the dictionary.

        Parameters
        ----------
        func : callable
            Function to apply.
        *args
            Positional arguments to pass to the function.
        **kwargs
            Keyword arguments to pass to the function.

        Returns
        -------
        BetterDict
            Result of applying the function to the keys of the object.
        """
        self.rename({key: func(key, *args, **kwargs) for key in self.keys()})
        return self

    def value_apply(self, func, *args, **kwargs):
        """Apply the function to each value of a dictionary.

        Parameters
        ----------
        func : callable
            Function to apply.
        *args
            Positional arguments to pass to the function.
        **kwargs
            Keyword arguments to pass to the function.

        Returns
        -------
        BetterDict
            Result of applying the function to each row of the object.
        """
        new_values = (func(value, *args, **kwargs) for value in self.values())
        for key, value in zip(self.keys(), new_values):
            self[key] = value
        return self


class BetterDict(dict, PicklerMixin, JoblibIOMixin, ApplyMixin):
    """
    Custom dictionary class that allows multiple get/set operations at once.

    Class also supports the following operations:

    - Access and set values using index notation.
    - Access keys referrencing the dictionary values.
    - Get the keys and values as lists.
    - Get the closest match to a given key.
    - Select a subset of the dictionary based on the values data types.
    - Apply functions row- and column-wise.
    - Perform I/O operations using ``pickle`` and ``joblib``.

    Examples
    --------
    Create a :class:`.BetterDict` instance from a normal dictionary:
    >>> d = BetterDict({"a": 1, "b": 2})
    >>> d["a"]
    1

    Get multiple values at once
    >>> d["a", "b"]
    {'a': 1, 'b': 2}

    Set multiple values at once
    >>> d["a", "b"] = 3, 4
    >>> d["a", "b"]
    {'a': 3, 'b': 4}

    Access values using index notation
    >>> d.iloc[0]
    3

    Set values using index notation
    >>> d.iloc[0, 1] = [5, 6]
    >>> d.iloc[:]
    [5, 6]

    Get the keys and values as lists
    >>> d.keys()    # Dictionary keys
    ['a', 'b']
    >>> d.values()  # Dictionary values
    [5, 6]

    Get the closest match to a given key
    >>> d.get_closest_match("A")
    'a'

    """

    @property
    def iloc(self) -> ILoc:
        """Access the dictionary values by index.

        Returns
        -------
        ILoc
            Dictionary values accessed by index.
        """
        return ILoc(self)

    @property
    def vloc(self) -> VLoc:
        """Access the dictionary values by value.

        Returns
        -------
        VLoc
            Dictionary values accessed by value.
        """
        return VLoc(self)

    @property
    def index(self) -> List[int]:
        """Get the indexes of the dictionary.

        Returns
        -------
        List[int]
            Indexes of the dictionary.
        """
        return self.iloc.index

    def __getitem__(
        self, key: Hashable | Iterable[Hashable],
    ) -> object | BetterDict:
        """Get the value(s) associated with the key(s).

        Parameters
        ----------
        key : Hashable | Iterable[Hashable]
            Key(s) to get the value(s) for.

        Returns
        -------
        object | BetterDict
            Value(s) associated with the key(s).
        """
        if iterable_not_string(key):
            return BetterDict({_key: self[_key] for _key in key})
        return self.get(key)

    def __setitem__(
        self, key: Hashable | Iterable[Hashable], value: Any | Iterable[Any],
    ):
        """Set the value(s) associated with the key(s).

        Parameters
        ----------
        key : Hashable | Iterable[Hashable]
            Key(s) to set the value(s) for.
        value : Any | Iterable[Any]
            Value(s) to set.
        """
        if same_length(key, value):
            for _key, _value in zip(key, value):
                super().__setitem__(_key, _value)
        else:
            super().__setitem__(key, value)

    def __getattr__(self, name: str) -> object:
        """Get the value associated with the key.

        Parameters
        ----------
        name : str
            Key to get the value for.

        Returns
        -------
        object
            Value associated with the key.
        """
        return self.__getitem__(name)

    def __setattr__(self, name: str, value: Any):
        """Set the value associated with the key.

        Parameters
        ----------
        name : str
            Key to set the value for.
        value : Any
            Value to set.
        """
        self.__setitem__(name, value)

    def rename(self, keys):
        """Rename the keys of the dictionary.

        Parameters
        ----------
        keys : dict
            New keys for the dictionary.
        """
        new_keys = keys.values()
        old_keys = keys.keys()
        for old_key, new_key in zip(old_keys, new_keys):
            value = self.pop(old_key)
            self[new_key] = value

    def keys(self) -> List[Hashable]:
        """Return a list of the dictionary keys.

        Returns
        -------
        List[Hashable]
            List of the dictionary keys.
        """
        return list(super().keys())

    def values(self) -> List[Any]:
        """Return a list of the dictionary values.

        Returns
        -------
        List[Any]
            List of the dictionary values.
        """
        return list(super().values())

    def close_match(self, key: Hashable, cutoff: float = 0.6) -> Hashable:
        """
        Return the key that is the closest match to the name from ``key``.

        Before applying the fuzzy match, the ``key`` and the dictionary
        keys are converted to lower case. This maximizes the chances of
        finding a match.

        Parameters
        ----------
        key : Hashable
            Key to find the closest match for.
        cutoff : float, default 0.6
            Minimum similarity ratio for a match to be returned. The parameter
            must be between 0 and 1. A value of 1 means that the strings must
            be identical. A value close to 0 means that the strings don't have
            to be identical, for a match to be returned.

        Returns
        -------
        Hashable
            Closest match to the given key.

        Raises
        ------
        KeyError
            If no match is found.
        """
        lower_keys = BetterDict(
            {str(_key).lower(): _key for _key in self.keys()}
        )
        lower_key = str(key).lower()
        matches = get_close_matches(  # type: ignore
            lower_key, lower_keys.keys(), cutoff=cutoff
        )
        if matches:
            return lower_keys[matches[0]]
        raise KeyError(f"No close match found for {key}.")

    def dtypes(self) -> BetterDict:
        """Get the data types of the dictionary values.

        Returns
        -------
        BetterDict
            Data types of the dictionary values.
        """
        return BetterDict({key: type(value) for key, value in self.items()})

    def select_dtypes(self, include=None, exclude=None):
        """
        Return a subset of the dictionary based on the data types.

        At least one of ``include`` or ``exclude`` must be specified and if
        both are specified, ``include`` cannot contain the same data types as
        ``exclude``.

        This function accepts some special strings to specify the data types:

        - ``'number'`` - Numeric data types.
        - ``'string'`` - String data types.
        - ``'pandas'`` - Pandas DataFrame or Series.
        - ``'numpy'`` - Numpy array.
        - ``'datetime'`` - Datetime data types.

        Parameters
        ----------
        include : list, default None
            Data types to include in the subset. If ``exclude`` is not
            specified, then this parameter is obligatory.
        exclude : list, default None
            Data types to exclude from the subset. If ``include`` is not
            specified, then this parameter is obligatory.

        Returns
        -------
        BetterDict
            Subset of the dictionary based on the data types.

        Raises
        ------
        ValueError
            - If both ``include`` and ``exclude`` are not specified.
            - If ``include`` contains the same data types as ``exclude``.

        Examples
        --------
        >>> from better_dict import BetterDict
        >>> d = BetterDict({'a': 1, 'b': '2', 'c': [3], 'd': {'e': 4}})
        >>> d.select_dtypes(include=['number', 'string'])
        BetterDict({'a': 1, 'b': '2'})

        Notes
        -----
        Method checks whether at least one of ``include`` or ``exclude`` is
        specified and if both are specified, it checks whether ``include``
        contains the same data types as ``exclude``. If any of these checks
        fail, a ``ValueError`` is raised.

        This method works similarly to the ``pandas.DataFrame.select_dtypes``
        method.

        See Also
        --------
        - :meth:`pandas.DataFrame.select_dtypes`
        - :func:`.translate_dtype`
        - :func:`better_dict.utils.make_list`
        - :func:`better_dict.utils.flatten`
        """
        if include is None and exclude is None:
            raise ValueError(
                "At least one of ``include`` or ``exclude`` must be specified."
            )
        include = tuple(
            flatten(
                [
                    translate_dtype(dtype) if isinstance(dtype, str) else dtype
                    for dtype in make_list(include)
                ]
            )
        )
        exclude_dtypes = tuple(
            flatten(
                [
                    translate_dtype(dtype) if isinstance(dtype, str) else dtype
                    for dtype in make_list(exclude)
                ]
            )
        )
        check_include_exclude = set(include).intersection(set(exclude_dtypes))
        if check_include_exclude:
            raise ValueError(
                "Cannot set the same data types to `include` and `exclude`:\n"
                "\n".join(f"- {dtype}" for dtype in check_include_exclude)
            )
        return BetterDict(
            {
                key: value
                for key, value in self.items()
                if isinstance(value, include)
                and not isinstance(value, exclude_dtypes)
            }
        )

    @classmethod
    def from_frame(cls, pandas_df: pd.DataFrame) -> BetterDict:
        """Create a ``BetterDict`` from a pandas DataFrame.

        Parameters
        ----------
        pandas_df : pd.DataFrame
            Pandas DataFrame to create the ``BetterDict`` from.

        Returns
        -------
        BetterDict
            ``BetterDict`` created from the pandas DataFrame.
        """
        return cls(pandas_df.to_dict(orient="list"))

    @classmethod
    def from_series(cls, pandas_series: pd.Series) -> BetterDict:
        """Create a ``BetterDict`` from a pandas Series.

        Parameters
        ----------
        pandas_series : pd.Series
            Pandas Series to create the ``BetterDict`` from.

        Returns
        -------
        BetterDict
            ``BetterDict`` created from the pandas Series.
        """
        return cls(pandas_series.to_dict())

    @classmethod
    def from_list(cls, list_obj: list) -> BetterDict:
        """Create a ``BetterDict`` from a list.

        Parameters
        ----------
        list_obj : list
            List to create the ``BetterDict`` from.

        Returns
        -------
        BetterDict
            ``BetterDict`` created from the list.
        """
        return cls(enumerate(list_obj))



"""
Accessor classes for :class:`better_dict.core.BetterDict` dictionary.

Classes
-------
This module defines the following classes:

- :class:`.VLoc`: Class allows the key retrieval using the dictionary values.
- :class:`.ILoc`: Class allows access and setting values using index notation.

Notes
-----
The :class:`.VLoc` and :class:`.ILoc` classes can be used by any dictionary-like
class. To create a new class that inherits from :class:`.VLoc` or :class:`.ILoc`
you must define the following properties:

- ``iloc``: property needed to implement the :class:`.ILoc` class.
- ``vloc``: property needed to implement the :class:`.VLoc` class.

The following example shows how to implement them:

.. code-block:: python

    class MyDict(dict, VLoc, ILoc):
        @property
        def iloc(self):
            return ILoc(self)

        @property
        def vloc(self):
            return VLoc(self)

In the example above, the ``MyDict`` class inherits from the ``dict``,
:class:`.ILoc` and :class:`.VLoc` classes. By defining the properties
``iloc`` and ``vloc`` and passing ``self`` to the :class:`.ILoc` and
:class:`.VLoc`, the dictionary keys and values can be accessed by the
:class:`.ILoc` and :class:`.VLoc` classes.

"""
from __future__ import annotations

from typing import Any, Hashable, Iterable, List

import numpy as np

from better_dict.utils import (iterable_not_string, same_length)


class VLoc:
    """
    Class for accessing the keys of a dictionary by value.

    Parameters
    ----------
    data : BetterDict
        Dictionary to access.

    Examples
    --------
    The following example shows how to use the :class:`.VLoc` class to access
    the keys of a dictionary by value.
    >>> from better_dict import BetterDict
    >>> d = BetterDict({'a': 1, 'b': 2, 'c': 3})
    >>> d.vloc[1]
    ['a']

    Notes
    -----
    The :class:`.VLoc` class only allows access to the keys of the dictionary
    by value. It does not allow manipulations of the dictionary keys or values.

    """

    def __init__(self, data):
        self.data = data

    def __getitem__(self, value: object | Iterable[object]) -> List[Hashable]:
        """Get the keys of the dictionary that match the value(s).

        Parameters
        ----------
        value : object | Iterable[object]
            Value(s) to match.

        Returns
        -------
        List[Hashable]
            keys(s) of the dictionary that match the value(s).
        """
        if not isinstance(value, (np.ndarray, list, tuple, set)):
            value = [value]
        return [_key for _key, _value in self.data.items() if
                any(v == _value for v in value)]

    def inverse(self):
        """Get the inverse of the dictionary.

        Returns
        -------
        BetterDict
            Inverse of the dictionary.
        """
        return self.data.__class__(
            {value: key for key, value in self.data.items()}
        )


class ILoc:
    """
    Class for manipulating :class:`.BetterDict` using index notation.

    Parameters
    ----------
    data : BetterDict
        Dictionary to access as a :class:`.BetterDict` object.

    Methods
    -------
    __getitem__(index: int | Iterable[int]) -> Any
        Get the value of the dictionary at the given index(es).
    __setitem__(index: int | Iterable[int], value: Any | Iterable[Any]) -> None
        Set the value(s) of the dictionary at the given index(es).

    Attributes
    ----------
    data : BetterDict
        Dictionary to access as a :class:`.BetterDict` object.

    Examples
    --------
    The following example shows how to use the :class:`.ILoc` class to access
    the values of a :class:`.BetterDict` dictionary using index notation:
    >>> from better_dict import BetterDict
    >>> d = BetterDict({'a': 1, 'b': 2, 'c': 3})
    >>> d.iloc[0]
    1

    You can also use the :class:`.ILoc` class get multiple values at once:
    >>> d.iloc[0, 1]
    [1, 2]
    >>> d.iloc[[0, 1]]
    [1, 2]
    >>> d.iloc[1:]
    [2, 3]

    Class allows setting multiple values at once using index notation:
    >>> d.iloc[[0, 1]] = [10, 20]
    >>> d
    {'a': 10, 'b': 20, 'c': 3}

    """

    def __init__(self, data):
        self.data = data

    def __getitem__(self, index: int | Iterable[int]) -> Any:
        """Get the value of the dictionary at the given index(es).

        Parameters
        ----------
        index : int | Iterable[int]
            Index(es) of the value(s) to get.

        Returns
        -------
        Any
            Value(s) of the dictionary at the given index(es).
        """
        values = list(self.data.values())
        return (
            [values[i] for i in index] if iterable_not_string(index)
            else values[index]
        )

    def __setitem__(
        self, index: int | Iterable[int], value: Any | Iterable[Any],
    ) -> None:
        """
        Set the value(s) of the dictionary at the given index(es).

        To set multiple values at once, you need to specify the same number of
        indices and values, for example:

        .. code-block:: python

            d.iloc[[0, 1]] = [10, 20]

        Parameters
        ----------
        index : int | Iterable[int]
            Index(es) of the value(s) to set.
        value : Any | Iterable[Any]
            Value(s) to set.

        """
        keys = list(self.data.keys())
        if same_length(index, value):
            for _index, _value in zip(index, value):
                self.data[keys[_index]] = _value
        else:
            self.data[keys[index]] = value

    @property
    def index(self) -> List[int]:
        """Get the indexes of the dictionary.

        Returns
        -------
        List[int]
            Indexes of the dictionary.
        """
        return list(range(len(self.data.keys())))



"""
Module defines helper functions for the ``better_dict`` package.

Functions
---------
- :func:`.iterable_not_string`: Check if the value is iterable but not a string.
- :func:`.same_length`: Check if the keys and values have the same length.
- :func:`.make_list`: Make a list from the value.
- :func:`.flatten`: Flatten the iterable object.

"""
from __future__ import annotations

from typing import Any, Generator, Iterable, List


__all__ = [
    "flatten",
    "iterable_not_string",
    "make_list",
    "same_length",
]


def iterable_not_string(value: object) -> bool:
    """Check if the value is iterable but not a string.

    Parameters
    ----------
    value : object
        Value to check.

    Returns
    -------
    bool
        ``True`` if the value is iterable but not a string, ``False`` otherwise.

    Examples
    --------
    >>> iterable_not_string([1, 2, 3])
    True
    >>> iterable_not_string('abc')
    False
    >>> iterable_not_string(1)
    False
    """
    return isinstance(value, Iterable) and not isinstance(value, str)


def same_length(keys: Any, values: Any) -> bool:
    """Check if the keys and values have the same length.

    Parameters
    ----------
    keys : Any
        Keys to check.
    values : Any
        Values to check.

    Returns
    -------
    bool
        True if the keys and values are iterables and have the same length,
        False otherwise.

    Examples
    --------
    >>> same_length([1, 2, 3], [4, 5, 6])
    True
    >>> same_length([1, 2, 3], [4, 5])
    False
    >>> same_length([1, 2, 3], 4)
    False
    """
    if all(
        iterable_not_string(value) and hasattr(value, "__len__")
        for value in [keys, values]
    ):
        return len(keys) == len(values)
    return False


def make_list(value: Any) -> List[Any]:
    """Make a list from the value.

    Parameters
    ----------
    value : Any
        Value to make a list from.

    Returns
    -------
    List[Any]
        List with the value.
    """
    if value is None:
        return []
    return list(value) if iterable_not_string(value) else [value]


def flatten(line: Any) -> Generator:
    """Flatten an arbitrarily nested sequence.

    Parameters
    ----------
    line : Any
        The possibly nested sequence to flatten.

    Returns
    -------
    flattened : generator
        A generator that yields the flattened sequence.

    Examples
    --------
    >>> list(flatten([[1, 2, 3], 4, [5, [6, 7, 8]]]))
    [1, 2, 3, 4, 5, 6, 7, 8]
    >>> list(flatten(1))
    [1]

    Notes
    -----
    - This function doesn't consider strings sequences.
    - If the input is not a sequence, it is returned as a single element list.
    """
    for element in make_list(line):
        if iterable_not_string(element):
            yield from flatten(element)
        else:
            yield element



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['better_dict']

package_data = \
{'': ['*']}

install_requires = \
['black>=21.10b0',
 'flake8-bandit>=2.1.2',
 'flake8-bugbear>=21.9.2',
 'flake8-docstrings>=1.6.0',
 'flake8-rst-docstrings>=0.2.5',
 'flake8>=4.0.1',
 'joblib>=1.2.0,<2.0.0',
 'numpy>=1.24.2,<2.0.0',
 'pandas>=1.5.3,<2.0.0',
 'pytest>=7.2.2,<8.0.0',
 'python-dotenv>=0.19.2,<0.20.0']

extras_require = \
{':python_full_version >= "3.6.2"': ['flake8-annotations>=2.7.0,<3.0.0']}

setup_kwargs = {
    'name': 'better-dict',
    'version': '0.0.1',
    'description': 'Python dictionary revamped.',
    'long_description': '![](docs/_static/EY_logo_5.gif)\n\n# Better Dict\n\n[![PyPI](https://img.shields.io/pypi/v/better-dict.svg)][pypi_]\n[![Status](https://img.shields.io/pypi/status/better-dict.svg)][status]\n[![Python Version](https://img.shields.io/pypi/pyversions/better-dict)][python version]\n[![License](https://img.shields.io/pypi/l/better-dict)][license]\n[![Read the documentation at https://better-dict.readthedocs.io/](https://img.shields.io/readthedocs/better-dict/latest.svg?label=Read%20the%20Docs)][read the docs]\n[![Codecov](https://codecov.io/gh/ingwersen-erik/better-dict/branch/main/graph/badge.svg)][codecov]\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit\\&logoColor=white)][pre-commit]\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)][black]\n\n[pypi_]: https://pypi.org/project/better-dict/\n\n[status]: https://pypi.org/project/better-dict/\n\n[python version]: https://pypi.org/project/better-dict\n\n[read the docs]: https://better-dict.readthedocs.io/\n\n[tests]: https://github.com/ingwersen-erik/better-dict/actions?workflow=Tests\n\n[codecov]: https://app.codecov.io/gh/ingwersen-erik/better-dict\n\n[pre-commit]: https://github.com/pre-commit/pre-commit\n\n[black]: https://github.com/psf/black\n\n## Description\n\nPython dictionary on steroids. The custom dictionary is inspired in\nby the functionalities that [pandas](https://pandas.pydata.org/) offers in\ntheir `DataFrame` and `Series` classes.\n\n***\n\n## Installation\n\nTo install **Better Dict**, execute the command:\n\n```console\n$ pip install better-dict\n```\n\n## Quickstart\n\nHere\'s a quick example of how to use **Better Dict**:\n\n```python\nimport better_dict as bd\n\nd = bd.BetterDict({"a": 1, "b": 2, "c": 3})\n\n# == Accessing values ==========================================================\n# Access multiple keys at once:\nd[["a", "b"]]  # returns {"a": 1, "b": 2}\n\n# Access dictionary values using item indexes:\nd.iloc[0]       # returns 1\nd.iloc[[0, 2]]  # returns [1, 3]\nd.iloc[1:]      # returns [2, 3]\n\n# Access dictionary keys using their values:\nd.vloc[1]       # returns "a"\nd.vloc[[1, 3]]  # returns ["a", "c"]\n\n# == Key Translations ==========================================================\n# Rename dictionary keys:\nd.rename({"a": "A", "b": "B", "c": "C"})    # returns {"A": 1, "B": 2, "C": 3}\n\n# == Apply Function ============================================================\n# Apply a function to all dictionary values:\nd.apply(lambda x: x + 1)                    # returns {"a": 2, "b": 3, "c": 4}\n\n# Apply a function to all dictionary keys:\nd.apply_keys(lambda x: x.upper(), axis=0)   # returns {"A": 1, "B": 2, "C": 3}\n\n# == I/O Operations ============================================================\n# Save dictionary to a Pickle file:\nd.to_pickle("d.pkl")\n\n# Load dictionary from a Pickle file:\nd = bd.BetterDict.from_pickle("d.pkl")\n\n# Save dictionary to a joblib file:\nd.to_joblib("d.joblib")\n\n# Load dictionary from a joblib file:\nd = bd.BetterDict.from_joblib("d.joblib")\n```\n\n## Q&A\n\n### 1. What is the ``BetterDict`` class and what additional functionality does it provide?\n\nThe ``BetterDict`` class is a custom subclass of Python\'s built-in dict class,\ndesigned to provide additional functionality for easier and more flexible\nmanipulation of dictionaries. The main enhancements include:\n\n* Accessing dictionary keys by value.\n* Manipulating dictionary keys and values using index notation.\n* Accessing and manipulating dictionary values using dot notation.\n* Other features include saving/loading dictionaries to/from files, creating\n  dictionaries from various data structures, applying functions to\n  dictionary values and keys, fuzzy key matching, and renaming dictionary keys.\n\n### 2. How can I access and set values in a ``BetterDict`` instance?\n\nAccessing and setting values in a ``BetterDict`` instance is made easy through a\nvariety of methods:\n\n* **``Get``/``Set`` values by key:** Use the standard dictionary syntax with square\nbrackets (e.g., ``d["key"]`` and ``d["key"] = value``).\n* **Get/Set multiple values at once:** Supply an iterable of keys\n  (e.g., ``d["key1", "key2"]`` and ``d["key1", "key2"] = value1, value2``).\n* **Index notation:** Use the iloc property to access and set values by index\n  (e.g., ``d.iloc[index]`` and ``d.iloc[index1, index2] = value1, value2``).\n* Additionally, dot notation can be used to access and set values (e.g., `d.key` and `d.key = value`).\n\n### 3. What are the available I/O operations for ``BetterDict`` and how can I use them?\n\n``BetterDict`` supports I/O operations using the **pickle** and **joblib** libraries,\nallowing you to easily save and load dictionaries to/from files. The main\nmethods for I/O operations are:\n\n* **Save with pickle:** Use the `save_pickle` method, supplying the file path\n  (e.g., `d.save_pickle("file_path.pkl")`).\n* **Load with pickle:** Use the `load_pickle` method, supplying the file path\n  (e.g., ``d = BetterDict.load_pickle("file_path.pkl"))``.\n* **Save with joblib:** Use the `save_joblib` method, supplying the file path\n  (e.g., `d.save_joblib("file_path.joblib")`).\n* **Load with joblib:** Use the `load_joblib` method, supplying the file path\n  (e.g., ``d = BetterDict.load_joblib("file_path.joblib")``).\n\n### 4. How can I create a ``BetterDict`` from different data structures like `pandas.DataFrame` or `numpy.ndarray`?\n\n``BetterDict`` offers class methods to create instances from various data\nstructures, such as pandas DataFrames, pandas Series, numpy arrays, and lists:\n\n- **From `pandas.DataFrame`:** Use the `from_frame` method (e.g., ``d = BetterDict.from_frame(data_frame))``.\n- **From `pandas.Series`:** Use the `from_series` method (e.g., ``d = BetterDict.from_series(data_series))``.\n- **From `numpy.ndarray`:** No direct method is available, but you can first convert the array to a pandas DataFrame and then use `from_frame`\n  (e.g., ``d = BetterDict.from_frame(pd.DataFrame(array)))``.\n- **From list:** Use the `from_list` method (e.g., ``d = BetterDict.from_list(list_obj))``.\n\nThese methods facilitate easy conversion between different data structures and ``BetterDict``.\n\n## Contributing\n\nContributions are welcome! If you have any suggestions or feature requests,\nplease open an issue or submit a pull request.\n\nFor more information on how to contribute to **Better Dict**,\nplease read the [Contributor Guide](./CONTRIBUTING.md).\n\n## License\n\nDistributed under the terms of the [MIT License](./LICENSE),\n*Better Dict* is free and open source software.\n\n<!-- github-only -->\n\n[license]: https://github.com/ingwersen-erik/better-dict/blob/main/LICENSE\n\n[contributor guide]: https://github.com/ingwersen-erik/better-dict/blob/main/CONTRIBUTING.md\n\n[command-line reference]: https://better-dict.readthedocs.io/en/latest/usage.html\n',
    'author': 'Erik Ingwersen',
    'author_email': 'erik.ingwersen@br.ey.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ingwersen-erik/better-dict',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



"""
Module contains classes that allow performing I/O operations on objects.

This module defines classes that serve as mixins for other classes.
These classes extensions inject methods that allow performing I/O
operations on objects.

Classes
-------
This module defines the following classes:

- :class:`.PicklerMixin`: Mixin class that adds methods for pickling objects.
- :class:`.JoblibIOMixin`: Mixin class that adds methods for saving objects as
  joblib files.

Examples
--------
The following example shows how to use the :class:`.PicklerMixin` class to
extend other classes, adding methods for pickling objects:

.. code-block:: python

    class MyClass(PicklerMixin):pass

    my_object = MyClass()
    # Save classe instance as pickle file
    my_object.to_pickle('my_object.pkl')

    # Load class instance from pickle file
    MyClass.from_pickle('my_object.pkl')

The same logic applies to the :class:`.JoblibIOMixin` class.

Notes
-----
All I/O classes from this module implement the saving method, using the ``to_``
prefix, and the loading method, using the ``from_`` prefix.
The loading methods are class methods, so they can be called directly from the
class, without the need to instantiate it first.
"""
from __future__ import annotations

import pickle
from pathlib import Path
from typing import Any, Dict

import joblib


class PicklerMixin:
    """
    Mixin class for pickle I/O.

    This class allows parent objects to perform I/O operations using pickle.
    The methods implemented in this class are:

    - :meth:`.PicklerMixin.to_pickle`: Save the object to a pickle file.
    - :meth:`.PicklerMixin.from_pickle`: Load the object from a pickle file.

    Examples
    --------
    >>> class MyClass(PicklerMixin):pass
    >>> obj = MyClass()

    >>> # To save the object to a pickle file:
    >>> obj.to_pickle("path/to/file")

    >>> # To load the object from a pickle file:
    >>> obj.from_pickle("path/to/file")

    """

    def __getstate__(self) -> Dict[str, Any]:
        """Get the state of the object for pickling.

        Returns
        -------
        Dict[str, Any]
            State of the object for pickling.
        """
        return self.__dict__

    def __setstate__(self, state: Dict[str, Any]):
        """Set the state of the object after unpickling.

        Parameters
        ----------
        state : Dict[str, Any]
            State of the object after unpickling.
        """
        self.__dict__ = state

    def to_pickle(self, path: str):
        """Save the object to a pickle file.

        Parameters
        ----------
        path : str
            Path to the pickle file.

        Raises
        ------
        ValueError
            If the path refers to a directory.
        """
        _path = Path(path).with_suffix(".pickle")
        if _path.is_dir():
            raise ValueError(f"{path!r} is a directory.")
        with open(_path, "wb") as file:
            pickle.dump(self, file)

    @classmethod
    def from_pickle(cls, path: str) -> PicklerMixin:
        """Load the object from a pickle file.

        Parameters
        ----------
        path : str
            Path to the pickle file.

        Returns
        -------
        PicklerMixin
            Object loaded from the pickle file.

        Raises
        ------
        ValueError
            If the path is not a file.
        """
        _path = Path(path).with_suffix(".pickle")
        if not _path.is_file():
            raise ValueError(f"{path!r} is not a file.")
        with open(_path, "rb") as file:
            # codiga-disable
            return cls(pickle.load(file))  # noqa


class JoblibIOMixin:
    """
    Mixin class for joblib I/O.

    This class allows parent objects to perform I/O operations using ``joblib``.
    The methods implemented in this class are:

    - :meth:`.JoblibIOMixin.to_joblib`: Save the object to a joblib file.
    - :meth:`.JoblibIOMixin.from_joblib`: Load the object from a joblib file.

    Methods
    -------
    to_joblib(path: str)
        Save the object to a joblib file.
    from_joblib(path: str)
        Load the object from a joblib file.

    Examples
    --------
    Define a class that inherits from :class:`.JoblibIOMixin`:
    >>> class MyClass(JoblibIOMixin):pass
    >>> obj = MyClass()

    Save the object as a ``joblib`` file:
    >>> obj.to_joblib("path/to/file")

    Load the object from a ``joblib`` file:
    >>> obj.from_joblib("path/to/file")

    """

    def to_joblib(self, path: str) -> None:
        """Save the object to a ``joblib`` file.

        Parameters
        ----------
        path : str
            Path to the ``joblib`` file.

        Raises
        ------
        ValueError
            If the path refers to a directory.
        """
        _path = Path(path).with_suffix(".joblib")
        if _path.is_dir():
            raise ValueError(f"{path!r} is a directory.")
        joblib.dump(self, _path)

    @classmethod
    def from_joblib(cls, path: str) -> JoblibIOMixin:
        """Load the object from a ``joblib`` file.

        Parameters
        ----------
        path : str
            Path to the ``joblib`` file.

        Returns
        -------
        JoblibIOMixin
            Object loaded from the ``joblib`` file.

        Raises
        ------
        ValueError
            If the path is not a file.
        """
        _path = Path(path).with_suffix(".joblib")
        if not _path.is_file():
            raise ValueError(f"{path!r} is not a file.")
        return cls(joblib.load(_path))  # noqa
